#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import copy
import logging

from spaceone.core import config
from spaceone.core.error import ERROR_CONFIGURATION, ERROR_UNKNOWN
from spaceone.core.scheduler.scheduler import IntervalScheduler
from spaceone.core.auth.jwt.jwt_util import JWTUtil
from spaceone.supervisor.service.supervisor_service import SupervisorService

_LOGGER = logging.getLogger(__name__)


__all__ = ['SyncScheduler']


def _get_domain_id_from_token(token):
    decoded_token = JWTUtil.unverified_decode(token)
    return decoded_token['did']


class SyncScheduler(IntervalScheduler):
    """ SyncScheduler
    """
    @staticmethod
    def get_task_metadata_and_params():
        try:
            name = config.get_global('NAME', '')
            hostname = config.get_global('HOSTNAME', '')
            token = config.get_global('TOKEN', '')

            if token == '':
                _LOGGER.error("TOKEN is not configured")
                raise ERROR_CONFIGURATION(key='TOKEN')

            if name == '':
                _LOGGER.error("name is not configured!")
                raise ERROR_CONFIGURATION(key='NAME')

            if hostname == '':
                _LOGGER.error("hostname is not configured!")
                raise ERROR_CONFIGURATION(key='HOSTNAME')

            tags = config.get_global('TAGS', {})
            labels = config.get_global('LABELS', {})
            domain_id = _get_domain_id_from_token(token)

            metadata = {
                'token': token,
                'domain_id': domain_id,
                'service': 'supervisor',
                'resource': 'Supervisor',
                'verb': 'sync_plugins'


            }

            params = {
                'name': name,
                'hostname': hostname,
                'tags': tags,
                'labels': labels,
                'domain_id': domain_id
            }

            return metadata, params

        except Exception as e:
            _LOGGER.error(f'[check_global_configuration] error: {e}', exc_info=True)
            raise ERROR_UNKNOWN(message=f'[check_global_configuration] error: {e}')

    def create_task(self):
        metadata, params = self.get_task_metadata_and_params()
        supervisor_svc: SupervisorService = SupervisorService(metadata)
        supervisor_svc.sync_plugins(copy.deepcopy(params))

        return []



#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ["KubernetesConnector"]

import logging
import time
from kubernetes import client
from kubernetes import config as k8s_config

from spaceone.core.error import ERROR_CONFIGURATION

from spaceone.supervisor.connector.container_connector import ContainerConnector

_LOGGER = logging.getLogger(__name__)

# max second for status checking
MAX_COUNT = 300
WAIT_CREATION = 30
ENDPOINT_INTERVAL = 10


class KubernetesConnector(ContainerConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _LOGGER.debug("[KubernetesConnector] config: %s" % self.config)
        self.headless = self.config.get('headless', False)
        self.node_selector = self.config.get('nodeSelector', {})
        self.NUM_OF_REPLICAS = 1
        self.namespace = self.config['namespace']

        try:
            k8s_config.load_incluster_config()
            conf = client.Configuration()
        except Exception as e:
            _LOGGER.debug(f'[KubernetesConnector] {e}')
            raise ERROR_CONFIGURATION(key='kubernetes configuration')

    def __del__(self):
        pass

    def search(self, filters):
        count = 0
        plugins_info = []
        # _LOGGER.debug("[KubernetesConnector] filters=%s" % filters)
        if 'label' in filters:
            services = self._list_services(filters['label'])
        else:
            services = []
        # _LOGGER.debug(services)
        count = len(services)
        for service in services:
            plugin = self._get_plugin_info_from_service(service)

            if self.headless:
                endpoints = plugin.get('endpoints', [])
                if len(endpoints) == 0:
                    continue

            plugins_info.append(plugin)

        return {'results': plugins_info, 'total_count': count}

    def run(self, image, labels, ports, name):
        """ Make sure, custom label is exist
        custom labels:
         - spaceone.supervisor.plugin_id
         - spaceone.supervisor.plugin.image
         - spaceone.supervisor.plugin.version
        """

        # ports (dict)
        # {'HostPort':80, 'TargetPort': 8080}
        # Docker API uses like
        # {'8080/tcp': 80}    , expose 8080/tcp to 80 (public)
        _LOGGER.debug(f'[run] create kubernetes deployment')

        resp_svc = self._get_service(labels, name, ports)
        resp_dep = self._get_deployment(labels, name, image)

        try:
            # Update endpoints, if needed
            # Wait a little
            # endpoints = self._update_endpoints(name)

            # _LOGGER.debug(f'[run] created deployment: {resp_dep}')
            plugin = self._get_plugin_info_from_service(resp_svc)

            # _LOGGER.debug(f'[run] plugin: {plugin}')
            return plugin

        except Exception as e:
            _LOGGER.error(f"[run] Failed to create kubernetes Service & Pod")
            _LOGGER.debug(e)
            raise ERROR_CONFIGURATION(key='kubernetes create')

    def stop(self, plugin):
        # TODO: seperated Service & Deployment
        try:
            name = plugin['name']
            # delete_namespaced_service
            k8s_core_v1 = client.CoreV1Api()
            resp_svc = k8s_core_v1.delete_namespaced_service(name, self.namespace)
            _LOGGER.debug(f'[stop] deleted service')

            # delete_namespaced_deployment
            k8s_apps_v1 = client.AppsV1Api()
            resp_dep = k8s_apps_v1.delete_namespaced_deployment(name, self.namespace)
            _LOGGER.debug(f'[stop] deleted deployment')

            return True
        except Exception as e:
            _LOGGER.error("Failed to stop docker")
            _LOGGER.debug(e)
            # TODO
            raise ERROR_CONFIGURATION(key='docker configuration')

    def _get_replica(self, service_type, plugin_id=None):
        _LOGGER.debug(f'[_get_replica] service_type: {service_type}, plugin_id: {plugin_id}')

        REPLICA_DIC = self.config.get('replica', {})
        service_type_with_plugin_id = service_type
        if plugin_id:
            service_type_with_plugin_id = f'{service_type}?{plugin_id}'

        if service_type_with_plugin_id in REPLICA_DIC:
            return REPLICA_DIC[service_type_with_plugin_id]
        elif service_type in REPLICA_DIC:
            return REPLICA_DIC[service_type]
        else:
            return self.NUM_OF_REPLICAS

    def _get_deployment(self, labels, name, image):
        """ Create or get Deployment

        Args:
            name: random generated name for service & deployment
        """
        k8s_apps_v1 = client.AppsV1Api()
        try:
            # get deployment
            resp_dep = k8s_apps_v1.read_namespaced_deployment(name=name, namespace=self.namespace)
            return resp_dep
        except Exception as e:
            _LOGGER.debug(f'[_get_deployment] may not found, {e}')

        # Create Deployment
        try:
            deployment = self._create_deployment(image, name, labels)
            resp_dep = k8s_apps_v1.create_namespaced_deployment(
                body=deployment, namespace=self.namespace)

            # create is asynchronous, wait a little
            time.sleep(WAIT_CREATION)

            # wait for max 5 minutes
            wait_count = 0
            while k8s_apps_v1.read_namespaced_deployment(name=name,
                                                         namespace=self.namespace).status.available_replicas < 1:
                time.sleep(10)
                wait_count += 10
                if wait_count > 300:
                    _LOGGER.debug(f'[_get_deployment] too much time to create deployment: {name}, {image}')

            return resp_dep
        except Exception as e:
            _LOGGER.debug(f'[_get_deployment] failed to create deployment, {e}')
            raise ERROR_CONFIGURATION(key='kubernetes configuration')

    def _create_deployment(self, image, name, labels):
        """ Create deployment content (dictionary)

        Args:
            param(dict): {
                    'name': ...

        Returns:
            deployment(dict)
        """
        mgmt_labels = self._get_k8s_label(labels)
        # _LOGGER.debug(f'mgmt_labels: {mgmt_labels}')
        if 'plugin_id' in mgmt_labels:
            plugin_id = mgmt_labels['plugin_id']
            NUM_OF_REPLICAS = self._get_replica(mgmt_labels['service_type'], plugin_id)
        else:
            NUM_OF_REPLICAS = self._get_replica(mgmt_labels['service_type'])

        deployment = {
            'apiVersion': 'apps/v1',
            'kind': 'Deployment',
            'metadata': {
                'name': name,
                'labels': mgmt_labels
            },
            'spec': {
                'replicas': NUM_OF_REPLICAS,
                'selector': {
                    'matchLabels': mgmt_labels
                },
                'template': {
                    'metadata': {
                        'name': name,
                        'labels': mgmt_labels
                    },
                    'spec': {
                        'containers': [
                            {
                                'image': image,
                                'name': name,
                                'imagePullPolicy': 'IfNotPresent',
                                'ports': [
                                    {
                                        'containerPort': 50051
                                    }
                                ],
                                'livenessProbe': {
                                    'tcpSocket': {
                                        'port': 50051
                                    }
                                }
                            }
                        ],
                        'nodeSelector': self.node_selector
                    }
                }
            }
        }

        if _service_account_name := self.config.get('service_account'):
            deployment['spec']['template']['spec']['serviceAccountName'] = _service_account_name

        if _image_pull_secrets := self.config.get('imagePullSecrets'):
            deployment['spec']['template']['spec']['imagePullSecrets'] = _image_pull_secrets

        if _volumes := self.config.get('volumes'):
            deployment['spec']['template']['spec']['volumes'] = _volumes

        if _container_env := self.config.get('env'):
            deployment['spec']['template']['spec']['containers'][0]['env'] = _container_env

        if _container_resources := self.config.get('resources'):
            deployment['spec']['template']['spec']['containers'][0]['resources'] = _container_resources

        if _container_volume_mounts := self.config.get('volume_mounts'):
            deployment['spec']['template']['spec']['containers'][0]['volumeMounts'] = _container_volume_mounts

        _LOGGER.debug(f'[_create_deployment] deployment: {deployment}')

        return deployment

    def _update_endpoints(self, svc_name):
        if self.headless is False:
            # Do nothing
            return
        endpoints = []
        wait_count = 0
        while wait_count < 300:
            endpoints = self._get_endpoints(svc_name)
            if endpoints != []:
                break
            time.sleep(10)
            wait_count += 10
        return endpoints

    def _get_service(self, label, name, port):
        """ Create or Get Service
        Return: service object
        """
        k8s_core_v1 = client.CoreV1Api()
        try:
            # get service
            resp_svc = k8s_core_v1.read_namespaced_service(name=name, namespace=self.namespace)
            # _LOGGER.debug(f'[run] found service: {resp_svc}')
            return resp_svc
        except Exception as e:
            _LOGGER.debug(f'[_get_service] may be not found, {e}')

        try:
            # Create Service
            service = self._create_service(label, name, port)
            # _LOGGER.debug(f'[run] service yml: {service}')
            resp_svc = k8s_core_v1.create_namespaced_service(
                body=service, namespace=self.namespace)
            # _LOGGER.debug(f'[run] created service: {resp_svc}')
            return resp_svc

        except Exception as e:
            _LOGGER.error(f"[run] Failed to create kubernetes Service")
            _LOGGER.debug(e)
            raise ERROR_CONFIGURATION(key='kubernetes configuration')

    def _create_service(self, labels, name, ports):
        """ Create Service content

        Returns:
            Service(dict)
        """
        # _LOGGER.debug(f'[_create_service] headless service: {self.headless}')
        mgmt_labels = self._get_k8s_label(labels)
        """  Example
            supervisor_name: root
            plugin_name: aws-ec2
            service_type: inventory.collector
            domain_id: domain-1234
        """
        if self.headless:
            spec = {
                'ports': [{
                    'port': ports['HostPort'], 'targetPort': ports['TargetPort']
                }],
                'selector': mgmt_labels,
                'clusterIP': 'None'
            }
        else:
            spec = {
                'ports': [{
                    'port': ports['HostPort'], 'targetPort': ports['TargetPort']
                }],
                'selector': mgmt_labels
            }

        service = {
            'apiVersion': 'v1',
            'kind': 'Service',
            'metadata': {
                'name': name,
                'annotations': labels,
                'labels': mgmt_labels
            },
            'spec': spec
        }
        return service

    def _list_services(self, label):
        """
        label(string):
            spaceone.supervisor.name=<supervisor name>
        label(list):
            ['spaceone.supervisor.name=<supervisor name>', 'a=b', ...]

        In K8S service, we put label at Annotation like

        Labels:            name: root
                           pluigin_id: plugin-885ff2c52a6c

        Annotations:       spaceone.supervisor.name: root
                           spaceone.supervisor.plugin.endpoint: grpc://root-plugin-885ff2c52a6c.dev-supervisor.svc.cluster.local:50051
                           spaceone.supervisor.plugin.image: pyengine/aws-ec2
                           spaceone.supervisor.plugin.version: 1.0
                           spaceone.supervisor.plugin_id: plugin-885ff2c52a6c
        """
        k8s_core_v1 = client.CoreV1Api()
        resp = k8s_core_v1.list_namespaced_service(
            namespace=self.namespace)
        result = []

        # labels
        labels = []
        if isinstance(label, list):
            labels = label
        else:
            labels = [label]

        # k,v = label.split("=")
        # _LOGGER.debug(f'[_list_service] labels: {labels}')
        for item in resp.items:
            if hasattr(item.metadata, 'annotations') and isinstance(item.metadata.annotations, dict):
                annotations = item.metadata.annotations
            else:
                annotations = {}
            if self._exist_label_in_annotation(labels, annotations):
                result.append(item)

        # _LOGGER.debug(f'[_list_service] services: {result}')
        return result

    def _get_endpoints(self, svc_name):
        """ This will be different from service type
        Headless Service: multiple endpoints
        Service: single endpoint
        """

        def _parse_subsets(response):
            subsets = response.subsets

            if subsets is None:
                _LOGGER.debug(f'[_get_endpoints] subsets is None : {response.metadata.name}')
                return []

            addrs = []
            port = None
            endpoints = []
            for subset in subsets:
                addrs = _parse_endpoints(subset.addresses)
                port = _parse_port(subset.ports)
            for addr in addrs:
                endpoint = f'grpc://{addr}:{port}'
                endpoints.append(endpoint)
            return endpoints

        def _parse_endpoints(addresses):
            """ Parse list of addresses
            """
            result = []
            for address in addresses:
                ip = address.ip
                result.append(ip)
            return result

        def _parse_port(ports):
            result = []
            for port in ports:
                svc_port = port.port
                result.append(svc_port)
            if len(result) == 1:
                return result[0]

        k8s_core_v1 = client.CoreV1Api()
        try:
            response = k8s_core_v1.read_namespaced_endpoints(
                name=svc_name,
                namespace=self.namespace)
            endpoints = _parse_subsets(response)
            # _LOGGER.debug(f'[_get_endpoints] {endpoints}')
            return endpoints
        except Exception as e:
            _LOGGER.error(f'[_get_endpoints] failed to get endpoints: {e}', exc_info=True)
            return []

    def _get_plugin_info_from_service(self, service):
        """
        service is V1Service object, not dictionary
        """
        # Custom Labels
        labels = service.metadata.annotations
        # _LOGGER.debug("[_get_plugin_info_from_service] labels=%s" % labels)

        if 'spaceone.supervisor.plugin_id' in labels:
            plugin_id = labels['spaceone.supervisor.plugin_id']
        else:
            plugin_id = "Unknown"
        if 'spaceone.supervisor.plugin.image' in labels:
            image = labels['spaceone.supervisor.plugin.image']
        else:
            image = "Unknown"
        if 'spaceone.supervisor.plugin.version' in labels:
            version = labels['spaceone.supervisor.plugin.version']
        else:
            version = "Unknown"
        if 'spaceone.supervisor.plugin.endpoint' in labels:
            endpoint = labels['spaceone.supervisor.plugin.endpoint']
        else:
            endpoint = "Unknown"

        plugin = {
            'plugin_id': plugin_id,
            'image': image,
            'version': version,
            'endpoint': endpoint,
            'labels': labels,
            'name': service.metadata.name,
            'status': self._update_state_machine(service.status)
        }

        if self.headless:
            endpoints = self._get_endpoints(service.metadata.name)
            plugin['endpoints'] = endpoints

        # _LOGGER.debug(f'[_get_plugin_info_from_service] plugin: {plugin}')
        return plugin

    @staticmethod
    def _exist_label_in_annotation(labels, annotation):
        """ Check existance of label in annotation
            (Exact match)

        Args:
            labels(list)
            annotation(dict)

        Return:
            True | False
        """
        result = False
        for label in labels:
            k, v = label.split("=")
            if k in annotation and annotation[k] == v:
                result = True
            else:
                # One of labels is not match
                return False
        return result

    @staticmethod
    def _get_k8s_label(labels):
        """ make OPS labels for K8S management
            labels = {
              spaceone.supervisor.name: root
              spaceone.supervisor.domain_id: domain-1234
              spaceone.supervisor.plugin.endpoint: grpc://random_string.dev-supervisor.svc.cluster.local:50051
              spaceone.supervisor.plugin.image: pyengine/aws-ec2
              spaceone.supervisor.plugin.version: 1.0
              spaceone.supervisor.plugin_id: plugin-885ff2c52a6c
            }

            Returns: mgmt_label (dict)
                {
                    supervisor_name: root
                    plugin_id: plugin-aws-ec2
                    domain_id: domain-1234
                }
        """
        mgmt_label = {}
        for k, v in labels.items():
            if k == 'spaceone.supervisor.name':
                mgmt_label['supervisor_name'] = v
            elif k == 'spaceone.supervisor.domain_id':
                mgmt_label['domain_id'] = v
            elif k == 'spaceone.supervisor.plugin_id':
                mgmt_label['plugin_id'] = v
            elif k == 'spaceone.supervisor.plugin.version':
                mgmt_label['version'] = v
            elif k == 'spaceone.supervisor.plugin.service_type':
                mgmt_label['service_type'] = v
        return mgmt_label

    @staticmethod
    def _update_state_machine(status):
        return "ACTIVE"



# -*- coding: utf-8 -*-
from spaceone.core.error import *


class ERROR_INSTALL_PLUGINS(ERROR_BASE):
    _message = 'install plugin failed: {plugins}'

class ERROR_DELETE_PLUGINS(ERROR_BASE):
    _message = 'delete plugin failed excluding: {plugins}'



#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ['SupervisorManager']

import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager

_LOGGER = logging.getLogger(__name__)


class SupervisorManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # container API backend
        self.backend = config.get_global('BACKEND')
        connectors_conf = config.get_global('CONNECTORS')
        plugin_conf = connectors_conf[self.backend]
        self.port_range = (plugin_conf['start_port'], plugin_conf['end_port'])

    def install_plugin(self, image_uri, labels, ports, name):
        """ Install Plugin
        """
        # determine connector name
        _LOGGER.debug(f'[install_plugin] image_uri: {image_uri}, labels: {labels}, ports: {ports}, name: {name}')
        #connector = self.locator.get_connector(self.backend, config=self.plugin_conf)
        connector = self.locator.get_connector(self.backend)
        r = connector.run(image_uri, labels, ports, name)
        return r

    def delete_plugin(self, plugin_id, version):
        """ Delete plugin
        """
        labels = [
            f'spaceone.supervisor.plugin_id={plugin_id}',
            f'spaceone.supervisor.plugin.version={version}'
        ]

        target_plugins = self.list_plugins_by_label(labels)
        _LOGGER.debug(f'[delete_plugin] labels: {labels}, target: {target_plugins}')
        # determine connector
        #connector = self.locator.get_connector(self.backend, config=self.plugin_conf)
        connector = self.locator.get_connector(self.backend)
        deleted_count = 0
        for plugin in target_plugins['results']:
            _LOGGER.debug(f'[delete_plugin] plugin: {plugin}')
            connector.stop(plugin)
            deleted_count += 1
        return deleted_count

    def create_endpoint(self, hostname):
        """ Determine endpoint of plugin
        """
        pass

    def list_plugins_by_label(self, label):
        """ Discover plugins based on label

        Args:
            label(string, label)
                - spaceone.supervisor.name=<supervisor name>
        """
        filters = {'label': label}
        try:
            #connector = self.locator.get_connector(self.backend, self.plugin_conf)
            connector = self.locator.get_connector(self.backend)
            data = connector.search(filters=filters)
            return data
        except Exception as e:
            _LOGGER.error("list_plugins_by_label: %s" % filters)
            _LOGGER.error(e)
            return {'total_count': 0, 'results': []}

    def get_plugin_from_repository(self, plugin_id, domain_id):
        """ Contact to repository service
        Find plugin_info

        """
        # Create Repository Connector
        connector = self.locator.get_connector("RepositoryConnector")
        plugin_info = connector.get_plugin(plugin_id, domain_id)
        return plugin_info

    def find_host_port(self):
        """ find host port for container port mapping
        """
        #connector = self.locator.get_connector(self.backend, config=self.plugin_conf)
        connector = self.locator.get_connector(self.backend)
        used_ports = connector.list_used_ports()
        _LOGGER.debug("Used ports list: %s" % used_ports)
        s, e = self.port_range
        host_ports = set(range(s, e))
        possible_ports = host_ports - used_ports
        _LOGGER.debug("Possible allocated port list: %s" % possible_ports)
        return possible_ports.pop()

    def get_plugin_endpoint(self, name, hostname, host_port):
        """ Find the GRPC endpoint of plugin

        Args:

        Return:
            - endpoint: grpc://abc.example.com:50051
        """
        if self.backend == 'DockerConnector':
            endpoint = f'grpc://{hostname}:{host_port}'
        elif self.backend == 'KubernetesConnector':
            endpoint = f'grpc://{name}.{hostname}:{host_port}'
        else:
            _LOGGER.error(f'[get_plugin_endpoint] undefined backend: {self.backend}')
        return endpoint



# -*- coding: utf-8 -*-
#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ["DockerConnector"]

import docker
import logging
import time

from spaceone.core.error import ERROR_CONFIGURATION

from spaceone.supervisor.connector.container_connector import ContainerConnector

_LOGGER = logging.getLogger(__name__)

# max second for status checking
MAX_COUNT = 180


class DockerConnector(ContainerConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _LOGGER.debug(f'[DockerConnector] config: {self.config}')
        try:
            self.client = docker.DockerClient(base_url='unix://var/run/docker.sock')
        except Exception as e:
            _LOGGER.debug(f'[DockerConnector] {e}')
            raise ERROR_CONFIGURATION(key='docker configuration')

    def __del__(self):
        self.client.close()

    def search(self, filters):
        count = 0
        plugins_info = []
        _LOGGER.debug(f'[search] filters: {filters}')
        containers = self.client.containers.list(filters=filters)
        count = len(containers)
        _LOGGER.debug(f'[search] discovered containers: {count}')
        for container in containers:
            # Custom Labels
            labels = container.labels
            _LOGGER.debug("[DockerConnector] labels=%s" % labels)
            if 'spaceone.supervisor.plugin_id' in labels:
                plugin_id = labels['spaceone.supervisor.plugin_id']
            else:
                plugin_id = "Unknown"
            if 'spaceone.supervisor.plugin.image' in labels:
                image = labels['spaceone.supervisor.plugin.image']
            else:
                image = "Unknown"
            if 'spaceone.supervisor.plugin.version' in labels:
                version = labels['spaceone.supervisor.plugin.version']
            else:
                version = "Unknown"
            if 'spaceone.supervisor.plugin.endpoint' in labels:
                endpoint = labels['spaceone.supervisor.plugin.endpoint']
            else:
                endpoint = "Unknown"

            # Ports
            ports = container.attrs['NetworkSettings']['Ports']
            plugin = {
                'docker_id': container.id,
                'plugin_id': plugin_id,
                'image': image,
                'version': version,
                'ports': ports,
                'endpoint': endpoint,
                'labels': container.labels,
                'name': container.name,
                'status': self._update_state_machine(container.status)
                }
            plugins_info.append(plugin)
        return {'results': plugins_info, 'total_count': count}

    def run(self, image, labels, ports, name):
        """ Make sure, custom label is exist
        custom labels:
         - spaceone.supervisor.plugin_id
         - spaceone.supervisor.plugin.image
         - spaceone.supervisor.plugin.version
        """

        # ports (dict)
        # {'HostPort':80, 'TargetPort': 8080}
        # Docker API uses like
        # {'8080/tcp': 80}    , expose 8080/tcp to 80 (public)
        docker_ports = {'%s/tcp' % ports['TargetPort']: int(ports['HostPort'])}
        # command = "/bin/bash -c 'sleep 360'"
        _LOGGER.debug("Create Docker ...")
        try:
            container = self.client.containers.run(image=image, labels=labels, ports=docker_ports,
                                                   name=name, detach=True, auto_remove=True)

            ######################
            # Wait until running
            ######################
            count = 1
            time.sleep(5)
            status = self._get_status(container.id)
            while status != "running":
                time.sleep(1)
                count = count + 1
                status = self._get_status(container.id)
                _LOGGER.debug(f'[run] docker status check: {status}')
                if count > MAX_COUNT:
                    break

            # Get up-to-date information
            container = self.client.containers.get(container.id)
            labels = container.labels
            if 'spaceone.supervisor.plugin_id' in labels:
                plugin_id = labels['spaceone.supervisor.plugin_id']
            else:
                plugin_id = "Unknown"
            if 'spaceone.supervisor.plugin.image' in labels:
                image = labels['spaceone.supervisor.plugin.image']
            else:
                image = "Unknown"
            if 'spaceone.supervisor.plugin.version' in labels:
                version = labels['spaceone.supervisor.plugin.version']
            else:
                version = "Unknown"

            # Ports
            ports = container.attrs['NetworkSettings']['Ports']

            plugin = {
                'docker_id': container.id,
                'plugin_id': plugin_id,
                'image': image,
                'version': version,
                'ports': ports,
                'labels': container.labels,
                'name': container.name,
                'status': self._update_state_machine(container.status)
                }

            return plugin

        except Exception as e:
            _LOGGER.error("Failed to run docker")
            _LOGGER.debug(e)
            # TODO: 
            raise ERROR_CONFIGURATION(key='docker configuration')

    def stop(self, plugin):
        container_id = plugin['docker_id']
        _LOGGER.debug(f'[docker stop] stop & delete {container_id}')
        try:
            container = self.client.containers.get(container_id)
            container.stop()
            container.remove(force=True)
            return True
        except Exception as e:
            _LOGGER.error("Failed to stop docker")
            _LOGGER.debug(e)
            # TODO
            raise ERROR_CONFIGURATION(key='docker configuration')

    def list_used_ports(self):
        """ Find used ports

        Returns:
            - set of port
        """
        containers = self.client.containers.list()
        allocated_ports = []
        for container in containers:
            if 'NetworkSettings' not in container.attrs:
                _LOGGER.debug(f'Skip {container} since no NetworkSettings')
                continue
            if 'Ports' not in container.attrs['NetworkSettings']:
                _LOGGER.debug(f'No Ports: {container}')
                continue

            ports = container.attrs['NetworkSettings']['Ports']
            """
            {'80/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '8111'}]}
            {'50051/tcp': None}
            """
            try:
                for (k, v) in ports.items():
                    if v is None:
                        continue
                    for host_map in v:
                        if 'HostPort' in host_map:
                            allocated_ports.append(int(host_map['HostPort']))
            except Exception as e:
                _LOGGER.error(f'Fail to parse ports information: {ports}, {e}')
                continue
        return set(allocated_ports)

    def _get_status(self, container_id):
        container = self.client.containers.get(container_id)
        return container.status

    def _update_state_machine(self, status):
        if status == 'running':
            return "ACTIVE"
        return "ERROR"



# -*- coding: utf-8 -*-

__all__ = ['PluginManagerInfo']

import functools
from spaceone.api.plugin.v1 import plugin_manager_pb2
from spaceone.core.pygrpc.message_type import change_struct_type, change_timestamp_type
from spaceone.plugin.model.plugin_model import PluginManager


def PluginManagerInfo(pm_vo: PluginManager):
    info = {
        'plugin_manager_id': pm_vo.plugin_manager_id,
        'name': pm_vo.name,
        'state': pm_vo.state,
        'is_public': pm_vo.is_public,
        'labels': change_struct_type(pm_vo.labels),
        'tags': change_struct_type(pm_vo.tags),
        'created_at': change_timestamp_type(pm_vo.created_at),
        'updated_at': change_timestamp_type(pm_vo.updated_at),
        'domain_id': pm_vo.domain_id
    }
    return plugin_manager_pb2.PluginManagerInfo(**info)




CACHES = {

}

CONNECTORS = {
    'PluginConnector': {
        # 'endpoint': {
        #    'v1': 'grpc://plugin:50051'
        # }
    },
    'RepositoryConnector': {
        # 'endpoint': {
        #    'v1': 'grpc://repository:50051'
        # }
    },
    'DockerConnector': {
        # "start_port": 50060,
        # "end_port": 50090
    },
    'KubernetesConnector': {
        # "start_port": 50060,
        # "end_port": 50090,
        # "namespace": "supervisor",
        # "service_account": "service_account_name",
        # "imagePullSecrets": [
        #     {"name": "credentials"}
        # ],
        # "env": [
        #     {"name": "HTTP_PROXY", "value": "http://***"},
        #     {...}
        # ],
        # "resources": {
        #     "requests": {
        #         "memory": "64Mi",
        #         "cpu": "250m"
        #     },
        #     "limits": {
        #         "memory": "128Mi",
        #         "cpu": "500m"
        #     }
        # },
        # "headless": True,
        # "replica": {
        #    "inventory.collector": 4
        # },
        # "nodeSelector": {
        #     "Category": "supervisor"
        # }
    }
}

HANDLERS = {
    # # TODO: add system key authentication handler
    # 'authentication': [{
    #     'backend': 'spaceone.core.handler.authentication_handler.AuthenticationGRPCHandler',
    #     'uri': 'grpc://identity:50051/v1/Domain/get_public_key'
    # }],
    # 'authorization': [{
    #    'backend': 'spaceone.core.handler.authorization_handler.AuthorizationGRPCHandler',
    #    'uri': 'grpc://identity:50051/v1/Authorization/verify'
    # }],
    # 'event': []
}


# Define Queue options
QUEUES = {
    # 'default_q': {
    #     'backend': 'spaceone.core.queue.redis_queue.RedisQueue',
    #     'host': 'redis',
    #     'port': 6379,
    #     'channel': 'supervisor_queue'
    # }
}

# Define scheduler options
SCHEDULERS = {
    # 'publish': {
    #     'backend': 'spaceone.supervisor.scheduler.publish_scheduler.PublishScheduler',
    #     'queue': 'default_q',
    #     'interval': 30
    # },
    # 'sync': {
    #     'backend': 'spaceone.supervisor.scheduler.sync_scheduler.SyncScheduler',
    #     'queue': 'default_q',
    #     'interval': 120
    # }
}

# This value should be global unique
# Update at config.yml
NAME = ""
HOSTNAME = ""
BACKEND = "DockerConnector"
# PLUGIN = {
#     "backend": "DockerConnector",
#     "start_port": 50060,
#     "end_port": 50090
# }

TAGS = {}
LABELS = {}

# This is admin user token for this domain
# If you want remote TOKEN for security, use TOKEN_INFO instead of TOKEN
TOKEN = ""
TOKEN_INFO = {}

# This is for unittest
ENDPOINTS = {}
LOG = {}




from spaceone.supervisor.connector.plugin_connector import PluginConnector
from spaceone.supervisor.connector.docker_connector import DockerConnector
from spaceone.supervisor.connector.kubernetes_connector import KubernetesConnector
from spaceone.supervisor.connector.repository_connector import RepositoryConnector



# -*- coding: utf-8 -*-
#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ["PluginConnector"]

import logging

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint
from spaceone.core.error import ERROR_WRONG_CONFIGURATION

_LOGGER = logging.getLogger(__name__)


class PluginConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _LOGGER.debug("config: %s" % self.config)
        if 'endpoint' not in self.config:
            raise ERROR_WRONG_CONFIGURATION(key='endpoint')

        if len(self.config['endpoint']) > 1:
            raise ERROR_WRONG_CONFIGURATION(key='too many endpoint')

        for (k, v) in self.config['endpoint'].items():
            # parse endpoint
            e = parse_endpoint(v)
            self.protocol = e['scheme']
            if self.protocol == 'grpc':
                # create grpc client
                self.client = pygrpc.client(endpoint="%s:%s" % (e['hostname'], e['port']), version=k)
            elif self.protocol == 'http':
                # TODO:
                pass

    def publish(self, param):
        """ Call Plugin.Supervisor.publish

        Returns: SupervisorInfo
        """
        # _LOGGER.debug(param)
        if self.protocol == 'grpc':
            # Try to connect via grpc client
            r = self.client.Supervisor.publish(param, metadata=self.transaction.get_connection_meta())
            return r
        elif self.protocol == 'http':
            raise ERROR_WRONG_CONFIGURATION(key=self.protocol)

    def list_plugins_by_hostname(self, supervisor_id, hostname, domain_id):

        if self.protocol == 'grpc':
            # Try to connect via grpc client
            param = {'domain_id': domain_id}
            if supervisor_id:
                param.update({'supervisor_id': supervisor_id})
            if hostname:
                param.update({'hostname': hostname})

            # _LOGGER.debug(f"param: {param}")
            plugins_info = self.client.Supervisor.list_plugins(param,
                                                               metadata=self.transaction.get_connection_meta())
            return plugins_info

        raise ERROR_WRONG_CONFIGURATION(key=self.protocol)



# -*- coding: utf-8 -*-
#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ['PluginServiceManager']

import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager

_LOGGER = logging.getLogger(__name__)


class PluginServiceManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def publish_supervisor(self, params):
        """ Get connector for plugin

        connector is gRPC client for Plugin Service
        """
        _LOGGER.debug("Manager:publish_supervisor")
        connector = self.locator.get_connector('PluginConnector')
        r = connector.publish(params)
        return r

    def list_plugins(self, supervisor_id, hostname, domain_id):
        """ Sync Plugins from Plugin Service
        """
        connector = self.locator.get_connector('PluginConnector')
        _LOGGER.debug("[supervisor_manager] list_plugin:%s" % hostname)
        r = connector.list_plugins_by_hostname(supervisor_id, hostname, domain_id)
        return r




#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
from setuptools import setup, find_packages

setup(
    name='spaceone-supervisor',
    version=os.environ.get('PACKAGE_VERSION'),
    description='SpaceONE supervisor service',
    long_description='',
    url='https://www.spaceone.dev/',
    author='MEGAZONE SpaceONE Team',
    author_email='admin@spaceone.dev',
    license='Apache Software License',
    packages=find_packages(),
    install_requires=[
        'spaceone-core',
        'spaceone-api',
        'docker',
        'kubernetes',
        'redis',
        'python-consul',
        'hashids'
    ],
    zip_safe=False,
)



import logging

from datetime import datetime
from google.protobuf.json_format import MessageToDict
from hashids import Hashids

from spaceone.core.error import ERROR_CONFIGURATION
from spaceone.core.service import *
from spaceone.core import config, cache
from spaceone.supervisor.error import ERROR_INSTALL_PLUGINS, ERROR_DELETE_PLUGINS
from spaceone.supervisor.manager.supervisor_manager import SupervisorManager
from spaceone.supervisor.manager.plugin_service_manager import PluginServiceManager

_LOGGER = logging.getLogger(__name__)

SUPERVISOR_SYNC_EXPIRE_TIME = 600


class SupervisorService(BaseService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._supervisor_mgr: SupervisorManager = self.locator.get_manager('SupervisorManager')
        self._plugin_service_mgr: PluginServiceManager = self.locator.get_manager('PluginServiceManager')

    @transaction
    @check_required(['name', 'hostname', 'domain_id'])
    def publish_supervisor(self, params):
        """ request publish
        The params comes from publish worker

        Args:
            params
              - name
              - hostname
              - tags
              - labels
              - secret_key
              - domain_id

        """
        # collect plugins_info
        result = self.discover_plugins(params['name'])
        plugins_info = result['results']
        count = result['total_count']
        # _LOGGER.debug(f'[publish_supervisor] plugins_info: {plugins_info}, count: {count}')
        _LOGGER.debug(f'[publish_supervisor] count: {count}')
        result = []
        for plugin_info in plugins_info:
            plugin = {
                'plugin_id': plugin_info['plugin_id'],
                'version': plugin_info['version'],
                'state': plugin_info['status'],
                'endpoint': plugin_info['endpoint']
            }
            if 'endpoints' in plugin_info:
                plugin['endpoints'] = plugin_info['endpoints']
            else:
                plugin['endpoints'] = [plugin_info['endpoint']]

            result.append(plugin)
        params2 = params.copy()
        params2['plugin_info'] = result

        # _LOGGER.debug(f'[publish_supervisor] params: {params2}')
        result_data = self._plugin_service_mgr.publish_supervisor(params2)
        return result_data

    @transaction
    @check_required(['hostname', 'name', 'domain_id'])
    def sync_plugins(self, params):
        """ Sync plugins from Plugin Service
        After sync, install plugins if not exist

        Args:
            params (dict): {
              'hostname': str,
              'name': str,
              'tags': dict,
              'labels': dict,
              'domain_id': str
            }
        """
        # Parameter check
        supervisor_id = params.get('supervisor_id', None)
        hostname = params.get('hostname', None)
        name = params.get('name', None)
        domain_id = params.get('domain_id', None)

        # LOCK (after next sync)
        # Drop if previous task is running
        lock = self._get_lock(domain_id, name)
        if lock:
            _LOGGER.debug(f'[sync_plugins] running ... drop this task')
            return False
        self._set_lock(domain_id, name)

        self._supervisor_mgr = self.locator.get_manager('SupervisorManager')
        if supervisor_id is None and hostname is None:
            self._release_lock(domain_id, name)
            raise ERROR_CONFIGURATION(key='supervisor_id | hostname')

        # list plugins from plugin service
        _LOGGER.debug("Find plugins at %s, %s" % (supervisor_id, hostname))
        try:
            plugins = self._plugin_service_mgr.list_plugins(supervisor_id, hostname, domain_id)
            num_of_plugins = plugins.total_count
            _LOGGER.debug(f'[sync_plugins] num of plugins: {num_of_plugins}')
        except Exception as e:
            _LOGGER.error(f'[sync_plugins] {e}')
            self._release_lock(domain_id, name)
            return False

        _LOGGER.debug(f'[sync_plugins] Check Plugin State')
        # if plugin state == RE_PROVISION, delete first
        try:
            self._check_plugin_state(plugins.results, params)
        except Exception as e:
            _LOGGER.error(f'[sync_plugins] fail to check plugins, {e}')

        _LOGGER.debug(f'[sync_plugins] Install Plugins')
        try:
            self._install_plugins(plugins.results, params)
        except Exception as e:
            _LOGGER.error(f'[sync_plugins] fail to install plugins, {e}')
            self._release_lock(domain_id, name)
            raise ERROR_INSTALL_PLUGINS(plugins)

        _LOGGER.debug(f'[sync_plugins] Clean up Plugins')
        try:
            self._delete_plugins(plugins.results, params)
        except Exception as e:
            _LOGGER.error(f'[sync_plugins] fail to delete plugins, {e}')
            self._release_lock(domain_id, name)
            raise ERROR_DELETE_PLUGINS(plugins=plugins)

        # Publish Again
        _LOGGER.debug(f'[sync_plugins] Publish Supervisor')
        try:
            self.publish_supervisor(params)
        except Exception as e:
            _LOGGER.debug(f'[sync_plugins] fail to public {e}')
            self._release_lock(domain_id, name)

        self._release_lock(domain_id, name)
        return True

    def _check_plugin_state(self, plugins, params):
        """ Check plugin state first
        if state == RE_PROVISIONING, delete plugin first
        """
        for plugin in plugins:
            dict_plugin = MessageToDict(plugin, preserving_proto_field_name=True)
            plugin_domain_id = dict_plugin['domain_id']
            dict_plugin.update(params)
            dict_plugin['domain_id'] = plugin_domain_id
            state = dict_plugin.get('state', None)
            # _LOGGER.debug(f'[_check_plugin_state] plugin_info: {dict_plugin}')
            if state == 'RE_PROVISIONING' or state == 'ERROR':
                # _LOGGER.debug(f'[_check_plugin_state] params: {params}')
                self.install_plugin(dict_plugin)
                delete_params = {
                    'plugin_id': dict_plugin['plugin_id'],
                    'version': dict_plugin['version'],
                    'domain_id': dict_plugin['domain_id']
                }
                self.delete_plugin(delete_params)

    def _install_plugins(self, plugins, params):
        """ Install plugin based on plugins

        Args:
            params (dict): {
              'hostname': str,
              'name': str,
              'domain_id': str
            }

        """
        for plugin in plugins:
            dict_plugin = MessageToDict(plugin, preserving_proto_field_name=True)
            _LOGGER.debug(f'[_install_plugins] dict_plugin: {dict_plugin}')
            plugin_domain_id = dict_plugin['domain_id']
            dict_plugin.update(params)
            dict_plugin['domain_id'] = plugin_domain_id
            # _LOGGER.debug(f'[_install_plugins] plugin_info: {dict_plugin}')
            if not self._exist_plugin(dict_plugin):
                # _LOGGER.debug(f'[_install_plugins] params: {params}')
                _LOGGER.debug(f'[_install_plugins] install_plugin: {dict_plugin}')
                self.install_plugin(dict_plugin)
                # _LOGGER.debug(f'[_install_plugins] installed: {params}')

    def _delete_plugins(self, plugins, params):
        """ Delete plugins excluding plugins
        """
        labels = [f'spaceone.supervisor.name={params["name"]}']
        current_plugins = self._supervisor_mgr.list_plugins_by_label(labels)
        for current_plugin in current_plugins['results']:
            if _is_members(current_plugin, plugins) is False:
                # _LOGGER.debug(f'[_delete_plugins] delete plugin: {current_plugin}')
                # Delete current_plugin
                delete_params = {
                    'plugin_id': current_plugin['plugin_id'],
                    'version': current_plugin['version']
                }
                self.delete_plugin(delete_params)
            else:
                _LOGGER.debug(f'[_delete_plugins] member plugin: {current_plugin}')

    def _exist_plugin(self, plugin):
        """ Find plugin at local
        """
        labels = [
            f'spaceone.supervisor.name={plugin["name"]}',
            f'spaceone.supervisor.plugin_id={plugin["plugin_id"]}',
            f'spaceone.supervisor.plugin.version={plugin["version"]}'
        ]
        plugins = self._supervisor_mgr.list_plugins_by_label(labels)
        # _LOGGER.debug(f'[_exist_plugin]\n {labels}\n{plugins}')
        if plugins['total_count'] > 0:
            return True
        return False

    @check_required(['name', 'plugin_id', 'version', 'hostname', 'domain_id'])
    def install_plugin(self, params):
        """ Install Plugin based on params

        Args:
            params (plugin_info):
              - name: name of supervisor
              - plugin_id: plugin ID
              - version
              - hostname : for updating plugin endpoint

        image is real uri from repository service, since we maintain multiple docker repository
        """
        # Find detailed plugin information 
        plugin_id = params['plugin_id']
        version = params['version']
        domain_id = params['domain_id']
        plugin_info = self._supervisor_mgr.get_plugin_from_repository(plugin_id, domain_id)
        # _LOGGER.debug(f'[install_plugin] plugin_info: {plugin_info}')
        # - image_uri
        # based on image, version, contact to repository API
        image_uri = "%s/%s:%s" % (
            plugin_info.registry_url,
            plugin_info.image, version)

        labels = {
            'spaceone.supervisor.name': params['name'],
            'spaceone.supervisor.plugin_id': params['plugin_id'],
            'spaceone.supervisor.domain_id': params['domain_id'],
            # 'spaceone.supervisor.plugin.plugin_name': plugin_info.name,
            'spaceone.supervisor.plugin.image': plugin_info.image,
            'spaceone.supervisor.plugin.version': params['version'],
            'spaceone.supervisor.plugin.service_type': plugin_info.service_type
        }
        # Determine port mapping
        host_port = self._supervisor_mgr.find_host_port()
        _LOGGER.debug("Choose Host Port: %d" % host_port)

        # ports(dict)
        # {'HostPort':80, 'TargetPort':80}
        # TODO: how to determine host port
        # TODO: target_port (how can we know?)
        target_port = 50051
        ports = {'HostPort': host_port, 'TargetPort': target_port}

        # container name(This will reduce duplicated container)
        name = f'{plugin_id}-{_create_unique_name()}'
        # Update plugin endpoint
        endpoint = self._supervisor_mgr.get_plugin_endpoint(name, params['hostname'], host_port)
        labels.update({'spaceone.supervisor.plugin.endpoint': endpoint})

        result_data = self._supervisor_mgr.install_plugin(image_uri, labels, ports, name)
        # _LOGGER.debug(f'[install_plugin] installed plugin info: {result_data}')
        # update endpoint
        return result_data

    @check_required(['plugin_id', 'version'])
    def delete_plugin(self, params):
        """ Delete Plugin

        Args:
            params(dict) = {
                'plugin_id': 'str',
                'version': 'str'
            }

        """
        result_data = self._supervisor_mgr.delete_plugin(
            params['plugin_id'],
            params['version'])
        # _LOGGER.debug(f'[delete_plugin] result: {result_data}')
        return result_data

    def discover_plugins(self, name):
        """ Discover plugins
        """
        label = f'spaceone.supervisor.name={name}'
        # _LOGGER.debug(f'[discover_plugins] label: {label}')
        plugins = self._supervisor_mgr.list_plugins_by_label(label)
        return plugins

    def _get_lock(self, domain_id, name):
        try:
            key = f"supervisor:{domain_id}:{name}"
            return cache.get(key)
        except Exception as e:
            _LOGGER.debug(f'[_get_lock] {key}, {e}')
            return False

    @staticmethod
    def _set_lock(domain_id, name):
        try:
            key = f"supervisor:{domain_id}:{name}"
            return cache.set(key, 1, expire=SUPERVISOR_SYNC_EXPIRE_TIME)
        except Exception as e:
            _LOGGER.debug(f'[_set_lock] {key}, {e}')
            return False

    @staticmethod
    def _release_lock(domain_id, name):
        try:
            key = f"supervisor:{domain_id}:{name}"
            return cache.delete(key)
        except Exception as e:
            _LOGGER.debug(f'[_release_lock] {key}, {e}')
            return False


def _is_members(plugin_info, plugins_vo):
    plugin_id = plugin_info['plugin_id']
    version = plugin_info['version']

    for plugin in plugins_vo:
        if plugin.plugin_id == plugin_id and plugin.version == version:
            return True
    return False


def _create_unique_name():
    """ Create random unique id for endpoint
    """
    hashids = Hashids(salt='_create_unique_name', alphabet='qwertyuioplkjhgfdsazxcvbnm')
    utcnow = datetime.utcnow()
    return hashids.encode(utcnow.year, utcnow.month, utcnow.day, utcnow.hour, utcnow.minute, utcnow.second)



# -*- coding: utf-8 -*-
#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint
from spaceone.core.error import ERROR_NOT_IMPLEMENTED


class ContainerConnector(BaseConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def search(self, filters):
        raise ERROR_NOT_IMPLEMENTED(name='search')

    def run(self, image, labels, ports, name):
        # Create Container
        raise ERROR_NOT_IMPLEMENTED(name='run')

    def stop(self, container_id):
        # Delete Container
        raise ERROR_NOT_IMPLEMENTED(name='stop')

    def get(self, container_id):
        raise ERROR_NOT_IMPLEMENTED(name='get')

    def list_used_ports(self):
        return set([])
